/* lmpparse.c
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <malloc.h>
#include <lmp.h>
#include <lmpmath.h>

/* Evaluate the top level expressions: braces or numbers
 *
 * Create new instance to evaluate the braced-in expression
 * and return the result.
 * If no braces are available, we will just return the next
 * number.
 * returns the read or computed value
 */
lmp_node_t * lmp_parse_braces () { 
  lmp_node_t * res;

  if (lmp_token_following_equals ("(")) {
    lmp_token_inc ();
    res = lmp_parse ();
    if (!lmp_token_following_equals (")"))
      lmp_error ("\")\" expected at argument %d\n", lmp_token_args ());
    lmp_token_inc ();

    return res; /* return inner value */
  }

  if (lmp_token_following_equals ("{")) {
    lmp_token_inc ();
    res = NEW_VAL (lmp_token_next_complex ());
    if (!lmp_token_following_equals ("}"))
      lmp_error ("\"}\" expected at argument %d\n", lmp_token_args ());
    lmp_token_inc ();
  }
  else if (lmp_token_next_is_number ())
    res = NEW_VAL (lmp_token_next_number ());
  else
    res = lmp_node_new (VAR, lmp_token_next ());

  return res; /* return next number */
}

/* Evaluate level 8 expressions
 *
 * This function parses and evaluates the binary operation
 * "^" (power).
 * It calls the top level evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_power () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_braces ();

  for (;;) {
    if (lmp_token_following_equals ("^"))
      op = LMP_POWER;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_logarithm ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}

lmp_node_t * lmp_parse_diff () {
  lmp_node_t * first;
  lmp_control op;

  for (;;) {
    if (lmp_token_following_equals ("diff"))
      op = LMP_DIFF;
    else if (lmp_token_following_equals ("int"))
      op = LMP_INT;
    else
      return lmp_parse_power ();

    lmp_token_inc ();
    first = lmp_parse_logarithm ();
    first = lmp_node_new (CONTROL, op, 2, first, lmp_parse_logarithm ());
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
    return first;
  }

  return first;
}

/* Evaluate level 7 expressions
 *
 * This function parses and evaluates unary operations
 * like sin, cos, lg, sqrt etc.
 * It calls the level 8 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_trigonometry () {
  lmp_node_t * operand;
  lmp_operator op;

  for (;;) {
    if (lmp_token_following_equals ("sin"))
      op = LMP_SINE;
    else if (lmp_token_following_equals ("cos"))
      op = LMP_COSINE;
    else if (lmp_token_following_equals ("tan"))
      op = LMP_TANGENT;
    else if (lmp_token_following_equals ("cot"))
      op = LMP_COTANGENT;
    else if (lmp_token_following_equals ("arcsin"))
      op = LMP_ARC_SINE;
    else if (lmp_token_following_equals ("arccos"))
      op = LMP_ARC_COSINE;
    else if (lmp_token_following_equals ("arctan"))
      op = LMP_ARC_TANGENT;
    else if (lmp_token_following_equals ("ln"))
      op = LMP_LOGARITHM;
    else if (lmp_token_following_equals ("lg"))
      op = LMP_LOG_10;
    else if (lmp_token_following_equals ("exp"))
      op = LMP_EXP;
    else if (lmp_token_following_equals ("sqrt"))
      op = LMP_SQRT;
    else if (lmp_token_following_equals ("-"))
      op = LMP_UNARY_MINUS;
    else
      return lmp_parse_diff ();

    lmp_token_inc ();
    operand = lmp_node_new (OP, op, 1, lmp_parse_logarithm ());
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
    return operand;
  }
}

/* Evaluate level 6 expressions
 *
 * This function parses and evaluates binary prefix
 * operation log. It calls the 7 level evaluation routine
 * for the arguments.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_logarithm () {
  lmp_node_t * first;
  lmp_operator op;

  for (;;) {
    if (lmp_token_following_equals ("log"))
      op = LMP_LOG;
    else
      return lmp_parse_trigonometry ();

    lmp_token_inc ();
    first = lmp_node_new (OP, op, 2, lmp_parse_logarithm (), lmp_parse_logarithm ());
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
    return first;
  }

  return first;
}


/* Evaluate level 5 expressions
 *
 * This function parses and evaluates the binary operations
 * mod (modulo) and div (integer div).
 * It calls the level 6 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_modulo () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_logarithm ();


  for (;;) {
    if (lmp_token_following_equals ("mod"))
      op = LMP_MODULO;
    else if (lmp_token_following_equals ("div"))
      op = LMP_DIV_NO_REM;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_logarithm ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}

/* Evaluate level 4 expressions
 *
 * This function parses and evaluates the binary operations
 * * and /.
 * It calls the level 5 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_mul () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_modulo ();


  for (;;) {
    if (lmp_token_following_equals ("*"))
      op = LMP_MULTIPLICATION;
    else if (lmp_token_following_equals ("/"))
      op = LMP_DIVISION;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_modulo ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}

lmp_node_t * lmp_parse_sum () {
  lmp_node_t * first, * second, * third;
  lmp_control op;

  for (;;) {
    if (lmp_token_following_equals ("sum"))
      op = LMP_SUM;
    else if (lmp_token_following_equals ("prod"))
      op = LMP_PROD;
    else
      return lmp_parse_mul ();

    lmp_token_inc ();
    first  = lmp_parse_logarithm ();
    second = lmp_parse_logarithm ();
    third  = lmp_parse_logarithm ();
    first  = lmp_node_new (CONTROL, op, 4, first, second, third, lmp_parse_logarithm ());
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());

    return first;
  }

  return first;
}

/* Evaluate level 3 expressions
 *
 * This function parses and evaluates the binary operations
 * + and -.
 * It calls the level 4 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_addition () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_sum ();


  for (;;) {
    if (lmp_token_following_equals ("+"))
      op = LMP_ADDITION;
    else if (lmp_token_following_equals ("-"))
      op = LMP_SUBTRACTION;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_sum ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}

/* Evaluate level 2 expressions
 *
 * This function parses and evaluates comparisons
 * like ==, < etc.
 * It calls the level 3 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_compare () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_addition ();


  for (;;) {
    if (lmp_token_following_equals ("=="))
      op = LMP_EQUALS;
    else if (lmp_token_following_equals ("!="))
      op = LMP_UNEQUALS;
    else if (lmp_token_following_equals (">="))
      op = LMP_GREATER_EQUAL;
    else if (lmp_token_following_equals ("<="))
      op = LMP_LOWER_EQUAL;
    else if (lmp_token_following_equals (">"))
      op = LMP_GREATER;
    else if (lmp_token_following_equals ("<"))
      op = LMP_LOWER;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_addition ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}

/* Evaluate level 1 expressions
 *
 * This function parses and evaluates logical operation
 * and.
 * It calls the level 2 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_node_t * lmp_parse_and () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_compare ();


  for (;;) {
    if (lmp_token_following_equals ("and"))
      op = LMP_OUTER_AND;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_compare ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}

/* Evaluate level 0 expressions
 *
 * This function parses and evaluates logical operation
 * "or" and calls the level 2 evaluation routine.
 * returns the computed value.
 */
lmp_node_t * lmp_parse () {
  lmp_node_t * left, * right;
  lmp_operator op;

  left = lmp_parse_and ();


  for (;;) {
    if (lmp_token_following_equals ("or"))
      op = LMP_OUTER_OR;
    else
      return left;

    lmp_token_inc ();
    right = lmp_parse_and ();
    left = lmp_node_new (OP, op, 2, left, right);
    if (lmp_proc_error != 0)
      lmp_stop (lmp_token_prev ());
  }

  return left;
}
